/*
 * Solve the four-color map problem for a particular map.
 * Also, solve related problems (the 8-color map on a torus?  no problem)
 * $Id: color.c,v 1.22 1997/08/21 21:21:56 kragen Exp kragen $
 */

#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

/*
 * Color stuff.
 * There are a finite number of colors, specified in the input file;
 * each color is assigned a number.
 * Color 0 is the 'unknown' color; that is, it is assigned to a country
 * when that country's color is unknown.
 * ncolors is one more than the maximum color number; that is, if we have
 * 'unknown', 'black', and 'white', ncolors is 3.
 */

int ncolors;
int allocated_colors;
char **colornames;
#define unknown_color 0

/*
 * Return the name of a particular color.
 */
char *
color(int i) {
	return colornames[i];
}

int 
nextcolor(int color) {
	color++;
	if (color == ncolors) {
		color = 0;
	}
	return color;
}

/*
 * These two functions return 0 on failure and nonzero on success.
 */

int
initcolors(int maxcolors) {
	assert(maxcolors > unknown_color);
	colornames = (char**)malloc(sizeof(int*) * maxcolors);
	if (colornames == (char**)NULL) return 0;

	/*
	 * This demonstrates that the only sensible choice for
	 * unknown_color is 0.
	 */
	ncolors = unknown_color + 1;
	colornames[unknown_color] = "unknown";
	
	allocated_colors = maxcolors;
	return 1;
}

int
addcolor(char *name) {
	if (ncolors == allocated_colors) return 0;
	colornames[ncolors] = strdup(name);
	if (colornames[ncolors] == (char*)NULL) return 0;
	ncolors++;
	return 1;
}

/*
 * There are some finite number of countries in the map; each is connected
 * to some number of neighbors.  No country can be the same color as any of
 * its neighbors.
 * Each country therefore has a name, a color, and a list of neighbors.
 * All countries are inside a single array, so they can refer to one another
 * by positions in the array. "My neighbors are countries 1, 3, and 17."
 */

typedef struct _intlist {
	int data;
	struct _intlist *next;
} intlist;

typedef struct _country {
	char *name;
	int color;
	intlist *neighbors;
} country;

/*
 * This function returns whether or not, in light of current knowledge, it's
 * OK for a country to be the color it is -- that is, if that color is shared
 * by none of its neighbors.
 */

int 
colorok(country countries[], int specific_country) {
	intlist *i;
	int color = countries[specific_country].color;

	if (color == unknown_color) {
		/*
		 * What to do in this situation is somewhat arbitrary.
		 * It's a "can't happen" the way I currently expect the
		 * code to run.
		 */
		assert(0);
		return 1;
	}

	for (i = countries[specific_country].neighbors; i != NULL; i = i->next) {
		if (countries[i->data].color == color) {
			return 0;
		}
		assert(i->data != specific_country);
	}
	return 1;
}

/*
 * This sets up a country for the first time.
 * It assumes that the name you pass it is in space that may be
 * overwritten.
 * It returns zero on failure and nonzero on success.
 */
int
init_country(country *p, char *name) {
	p->neighbors = (intlist*) NULL;
	p->name = strdup(name);
	if (p->name == (char*)NULL) return 0;
	p->color = unknown_color;
	return 1;
}

/*
 * This function returns 0 on failure, or nonzero on success.
 */
int
add_neighbor(country *p, int neighbor) {
	intlist *ilp = (intlist*)malloc(sizeof(intlist));
	if (ilp == (intlist*)NULL) return 0;
	ilp->data = neighbor;
	ilp->next = p->neighbors;
	p->neighbors = ilp;
	return 1;
}



typedef struct _map {
	country *countries;
	int ncountries;
	int countries_allocated;
} map;

/*
 * This function rolls into one the allocation and the initialization of
 * a new map.  It could be factored if needed.
 */

map *
new_map(int size) {
	map *p = (map*)malloc(sizeof(map));
	if (p == (map*)NULL) return (map*)NULL;

	p->countries = (country*)malloc(sizeof(country)*size);
	if (p->countries == (country*)NULL) {
		free(p);
		return (map*)NULL;
	}

	p->ncountries = 0;
	p->countries_allocated = size;

	return p;
}

/*
 * This finds the country with a particular name.
 * If it doesn't exist, it will create one.
 * The creation functionality could (and maybe should) be factored.
 * It returns the index into the countries array, or -1 on error.
 */
int
country_called(map *m, char *name) {
	int i;
	country *countries = m->countries;

	for (i = 0; i != m->ncountries; i++) {
		if (strcmp(name, countries[i].name) == 0) {
			return i;
		}
	}

	/* Creation begins here */
	/*
	 * This is a little inflexible; ideally, we'd reallocate the whole
	 * countries array.  Maybe later.
	 */
	if (m->ncountries == m->countries_allocated) return -1;

	/*
	 * This line is superfluous, until someone changes the way the
	 * loop above works, or inserts code in between.
	 */
	i = m->ncountries;
	if (init_country(&countries[i], name) == 0) {
		return -1;
	}
	m->ncountries++;
	return i;
}


/* 
 * This connects two countries by name.
 * It expects names because that's how I expect the input data to be.
 * Since it calls country_called, it will create any necessary countries.
 * It returns 0 on failure or nonzero on success.
 */
int
connect_countries(map *m, char *name1, char *name2) {
	int country1, country2;
	country *countries;
	country1 = country_called(m, name1);
	country2 = country_called(m, name2);
	if ((country1 == -1) || (country2 == -1)) return 0;

	countries = m->countries;
	if (add_neighbor(&countries[country1], country2) == 0) return 0;
	if (add_neighbor(&countries[country2], country1) == 0) return 0;
	return 1;
}


int
print_map (map *m) {
	int i;

	for (i = 0; i != m->ncountries; i++) {
		printf("%s %s\n", m->countries[i].name, color(m->countries[i].color));
	}
	return 1;
}

/*
 * Here's the algorithm itself.
 * You give it a map with all the countries colored unknown_color.
 * It either colors the map or leaves it all unknown.
 * If it colors the map, it returns nonzero.
 */

int
color_map(map *m) {
	int cc = 0;  /* Current country */
	country *countries = m->countries;
	country *ccp; /* Current country pointer */

	/*
	 * We know we've colored all the countries when the
	 * current-country index moves past the end of the list of
	 * countries.
	 * At any time during this loop, all countries less than the cc
	 * are colored, and all countries greater than the cc are
	 * uncolored.
	 */
	while (cc != m->ncountries && cc != -1) {
		ccp = &(countries[cc]);
		ccp->color = nextcolor(ccp->color);
		if (ccp->color == unknown_color) {
			printf("Giving up at %s\n", ccp->name);
			print_map(m);
			/*
			 * This condition means we've cycled through
			 * all possible colors for this country, and none
			 * of them worked, so we'll try changing some
			 * earlier countries.
			 */
			cc--;
		} else if (colorok(countries, cc)) {
			/*
			 * Well, if this color is OK based on what we
			 * know now, let's see if we can color some other
			 * countries while this country is this color.
			 * If it doesn't work out, we can come back and
			 * change this ocuntry later.
			 */
			cc++;
		} else {
			/*
			 * This color won't work.  Try another one for
			 * this country.
			 */
		}
	}
	/*
	 * If we're at the bottom of the loop, then one of the termination
	 * conditions was met.  Either cc is -1, meaning we couldn't find
	 * a color that would work for country 0 (no matter what), or
	 * cc is m->ncountries, in which case we won.
	 */
	if (cc == -1) return 0;
	assert(cc == m->ncountries);
	return 1;
}


/*
 * Well, it would be nice to be able to read in a map from a file.  Like,
 * for instance, a text file.  All we need to do is declare the colors
 * to be used and list pairs of countries.
 * So, here's the format.
 * A line beginning with # will be ignored.
 * A line whose first word is colors: lists colors to add.
 * Any other line with two words on it lists two countries that are
 * connected.
 */

/*
 * This finds the next word in the input string.
 * It tries to copy it into the buf; it always null-terminates the buf.
 * This whole deal is so much easier when strings are passed around as a
 * pointer and a count -- I don't have to copy the word!  Just get a pointer
 * to its beginning, and its length...
 * It returns a pointer to the input character after the last one it copied,
 * which will be part of the word if it ran out of buffer space, the space
 * after the word if it found a word and there are maybe more, or the null
 * at the end of the string if it got to the end of the string.
 * If there is whitespace after the last word in a string, calling this
 * function repeatedly will eventually result in a zero-length word.
 */

char *
copy_word(char *string, char *buf, int buflen) {
	char *sp = string, *bp = buf;
	assert(buflen != 0);
	while (*sp && isspace(*sp)) sp++;

	/*
	 * we do this to leave space for the null at the end of the buf.
	 */
	buflen--;

	while (*sp && !isspace(*sp) && buflen) {
		*bp = *sp;
		sp++;
		bp++;
		buflen--;
	}
	
	*bp = '\0';
	return sp;
}

/*
 * This function reads in a map from a stdio file.
 * The format is as described in the comments above copy_word.
 */

#define inbuflen 1023
#define wordlen 255
map *
read_file(FILE *input, int maxcountries, int maxcolors) {
	char inbuf[inbuflen];
	char word[wordlen];
	char otherword[wordlen];
	char *wp;
	map *m;

	if (initcolors(maxcolors) == 0) return (map*)NULL;

	m = new_map(maxcountries);
	if (m == (map*)NULL) return (map*)NULL;
	

	while(fgets(inbuf, inbuflen, input) != NULL) {
		if (inbuf[0] == '#') continue;

		wp = inbuf;
		wp = copy_word(wp, word, wordlen);

		if (strlen(word) == 0) continue;

		if (strcmp(word, "colors:") == 0) {
			while (*wp != '\0') {
				wp = copy_word(wp, word, wordlen);
				if (strlen(word) != 0) {
					addcolor(word);
				}
			}
		} else {
			strcpy(otherword, word);
			wp = copy_word(wp, word, wordlen);
			if (strlen(word) == 0) {
				fprintf(stderr, "Invalid input line %s", inbuf);
				continue;
			}
			if (!connect_countries(m, otherword, word)) {
				fprintf(stderr, 
					"connect_countries failed on %s %s\n",
					otherword, word);
			}
		}
	}

	return m;
}

/*
 * main for real running
 */
int
main(int argc, char **argv) {
	map *the_map;
	FILE *infile;

	infile = fopen(argv[1], "r");
	if (infile == (FILE*)NULL) { 
		perror("fopen");
		exit(1);
	}

	the_map = read_file(infile, 1023, 127);

	if (color_map(the_map)) {
		printf("\n");
		print_map(the_map);
	} else {
		printf("Couldn't color the map\n");
	}

	return 0;
}
